﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Windows.Forms;
using FTD2XX_NET; //Aggiungere riferimento libreria FTDI2XX_NET
using System.Speech.Synthesis;
using System.Threading;
using System.Collections.Generic;
using System.IO;


namespace CLI_GUI
{
    public partial class Form1 : Form
    {
        FTDI myFtdiDevice;
        FTDI.FT_STATUS ftStatus = FTDI.FT_STATUS.FT_OK;
        AutoResetEvent ReceivedDataEvent = new AutoResetEvent(false);
        // Initialize a new instance of the SpeechSynthesizer.
        SpeechSynthesizer synth = new SpeechSynthesizer();
        //System.Timers.Timer tmrFTDI;
        System.Windows.Forms.Timer tmrFTDI;
        System.Timers.Timer tmrSquilli;
        private uint nSquilli = 0;
        byte[] RxMsg = null;
        delegate void MyDelegate(int StatoMsg, byte[] MsgCLI);
        string Annuncio = null;
        bool ChiamataAnonima = false;

        Dictionary<string, string> rubrica = new Dictionary<string, string>();

        //-------------------------------------------------------------------------------------------------------------------------------------------
        // Costruttore della classe Form1
        //-------------------------------------------------------------------------------------------------------------------------------------------
        public Form1()
        {
            InitializeComponent();

            tmrSquilli = new System.Timers.Timer();
            tmrSquilli.Interval = 7000; // in mmSec
            tmrSquilli.Elapsed += new System.Timers.ElapsedEventHandler(this.tmrSquilli_Tick);
        }


        //-------------------------------------------------------------------------------------------------------------
        // Tipi di eventi e mappatura bit modem status e Line Status (estratti dal wrapped dll)
        //-------------------------------------------------------------------------------------------------------------
        /* FTDI device event types that can be monitored
        public class FT_EVENTS
            public const UInt32 FT_EVENT_RXCHAR = 0x00000001;
            public const UInt32 FT_EVENT_MODEM_STATUS = 0x00000002;
            public const UInt32 FT_EVENT_LINE_STATUS = 0x00000004;
        The modem status is bit-mapped as follows: 
            Clear To Send(CTS) = 0x10, 
            Data Set Ready(DSR) = 0x20,
            Ring Indicator(RI) = 0x40, 
            Data Carrier Detect(DCD) = 0x80.
        The line status is bit-mapped as follows: 
            Overrun Error(OE) = 0x02,
            Parity Error(PE) = 0x04,
            Framing Error(FE) = 0x08,
            Break Interrupt(BI) = 0x10.
        */
        //-------------------------------------------------------------------------------------------------------------
        // Thread di attesa dati seriali in arrivo
        //-------------------------------------------------------------------------------------------------------------
        private void ReadData(object pSender, DoWorkEventArgs pEventArgs)
        {
            //FTDI.FT_STATUS status = FTDI.FT_STATUS.FT_OK;
            //BackgroundWorker worker = (BackgroundWorker)pSender;
            uint nrOfBytesAvailable = 0;
            uint nByteRead = 0;
            uint sMsg = 0;
            uint nByteMsg = 0;
            uint k = 0;
            uint TipoEvento = 0;
            uint CheckSum = 0;
            byte StatoModem = 0;
            bool CliMsgRic = false;
            //-------------------------------------------------------------------------------------------------------------------------------------------
            // Attiva evento cambiamento linee modem status ed entra in attesa del primo squillo telefonico
            //-------------------------------------------------------------------------------------------------------------------------------------------
            myFtdiDevice.SetEventNotification(FTDI.FT_EVENTS.FT_EVENT_MODEM_STATUS,    //Attiva evento modem Status (CTS)
                                              ReceivedDataEvent);
            while(true) {
                //In attesa del primo squillo telefonico
                ReceivedDataEvent.WaitOne(); //In attesa di un evento da FTDI (CTS low)
                myFtdiDevice.GetEventType(ref TipoEvento);
                if (TipoEvento == FTDI.FT_EVENTS.FT_EVENT_MODEM_STATUS)
                {
                    // ------ Evento cambiamento linee modem (CTS) ------
                    myFtdiDevice.GetModemStatus(ref StatoModem);
                    if ((StatoModem & FTDI.FT_MODEM_STATUS.FT_CTS) == 0)  //Linea CTS=1 -> (FT_CTS=0) Rilevato inizio squillo telefonico (Doppio MOSFET)
                    {  
                        tmrSquilli.Stop(); //Start timer per conteggio squilli
                        if (nSquilli == 0) //Primo squillo
                        { 
                          myFtdiDevice.SetDTR(true); // DTR=0 -> Attiva MT88E39
                          myFtdiDevice.SetEventNotification(FTDI.FT_EVENTS.FT_EVENT_RXCHAR          //Attiva evento dati seriali
                                                            |FTDI.FT_EVENTS.FT_EVENT_MODEM_STATUS,   //Attiva evento modem Status (CTS)
                                                             ReceivedDataEvent);
                          
                          CliMsgRic = false;
                          sMsg = 0;
                        }
                        else //Squilli successivi
                        { 
                            myFtdiDevice.SetDTR(false); // DTR=1 -> Disattiva MT88E39
                            myFtdiDevice.SetEventNotification(FTDI.FT_EVENTS.FT_EVENT_MODEM_STATUS,   //Attiva solo evento modem status
                                                              ReceivedDataEvent);
                            myFtdiDevice.Purge(FTDI.FT_PURGE.FT_PURGE_RX);  //purge buffere rx
                            if (CliMsgRic) {
                                if (ChiamataAnonima) {
                                    synth.SpeakAsync(Convert.ToString(4 - nSquilli));
                                    if (nSquilli == 4) {
                                        myFtdiDevice.SetRTS(true); //Impegna la linea per 10 secondi
                                        tmrSquilli.Interval = 10000; // in mmSec
                                    }
                                } else if ((nSquilli % 2) == 0) synth.SpeakAsync(Annuncio); //Ripeti annuncio ogni 2 squilli
                            }
                            else BeginInvoke(new MyDelegate(MessaggioCLIArrivato), new object[] { 2, RxMsg });
                        }
                        nSquilli++;
                        myFtdiDevice.Purge(FTDI.FT_PURGE.FT_PURGE_RX);  //purge buffere rx
                        tmrSquilli.Start();
                    }
                }
                else if (TipoEvento == FTDI.FT_EVENTS.FT_EVENT_RXCHAR)
                {
                    // ------ Evento ricezione byte da seriale --------
                    myFtdiDevice.GetRxBytesAvailable(ref nrOfBytesAvailable);
                    while (nrOfBytesAvailable > 0)
                    {
                        byte[] readData = new byte[nrOfBytesAvailable];
                        myFtdiDevice.Read(readData, nrOfBytesAvailable, ref nByteRead); //54 byte del CLI letti

                        for (int i = 0; i < nByteRead; i++) // Analizza i Byte letti
                        { 
                            switch (sMsg)
                            {
                                case 0: //Attesa codice 0x80
                                    if (readData[i] == 0x80)
                                    {
                                        sMsg++;
                                        CheckSum = readData[i];
                                    }
                                    break;
                                case 1: //Lettura numero byte che seguono esclusa la checksum
                                    nByteMsg = readData[i];
                                    sMsg++;
                                    k = 0;
                                    RxMsg = new byte[nByteMsg];
                                    CheckSum += readData[i];
                                    break;
                                case 2: //Lettura messaggio
                                    RxMsg[k++] = readData[i];
                                    if (k == nByteMsg) sMsg++;
                                    CheckSum += readData[i];
                                    break;
                                case 3: //Controllo checkSum
                                    CheckSum += readData[i];
                                    if ((CheckSum & 0xFF) == 0)
                                    {
                                        BeginInvoke(new MyDelegate(MessaggioCLIArrivato), new object[] { 0, RxMsg });
                                    }
                                    else
                                    {
                                        BeginInvoke(new MyDelegate(MessaggioCLIArrivato), new object[] { 1, RxMsg });
                                    }
                                    sMsg = 0;
                                    CliMsgRic = true;
                                    myFtdiDevice.Purge(FTDI.FT_PURGE.FT_PURGE_RX);  //purge buffere rx
                                    break;
                            }
                        }
                        myFtdiDevice.GetRxBytesAvailable(ref nrOfBytesAvailable);
                    } // Loop fintanto che ci sono byte disponibili dalla seriale FTDI
                }
            } // fine ReadData
        }

        //-------------------------------------------------------------------------------------------------------------------------------------------
        // Scadenza del timer tmrSquilli
        //-------------------------------------------------------------------------------------------------------------------------------------------
        public void tmrSquilli_Tick(object sender, System.Timers.ElapsedEventArgs arg)
        {
            tmrSquilli.Stop();
            //ScriviSuMonitor("Numero Squilli: " + nSquilli); //non funziona perchè se è da un thread diverso
            //synth.SpeakAsync("Numero squilli: " + nSquilli); //funziona ma è inutile
            RxMsg = new byte[1];
            RxMsg[0] = (byte)nSquilli;
            BeginInvoke(new MyDelegate(MessaggioCLIArrivato), new object[] { 3, RxMsg });
            myFtdiDevice.SetDTR(false); //Disattiva MT88E39
            if (ChiamataAnonima)
            {
                if (nSquilli >= 4) synth.SpeakAsync("Chiamata rifiutata");
                ChiamataAnonima = false;
                myFtdiDevice.SetRTS(false); //Rilascia la linea telefonica
            }
            myFtdiDevice.SetEventNotification(FTDI.FT_EVENTS.FT_EVENT_MODEM_STATUS,
                                              ReceivedDataEvent);
            nSquilli = 0;
        }

        //-------------------------------------------------------------------------------------------------------------------------------------------
        // Messaggio CLI Arrivato 
        // StatoMsg = 0 : OK
        // StatoMsg = 1 : Errore di CheckSum
        // StatoMsg = 2 : Rilevati dati seriali non CLI tra il primo squillo e il secondo  (MsgCLI = null)
        //-------------------------------------------------------------------------------------------------------------------------------------------
        void MessaggioCLIArrivato(int StatoMSg, byte[] MsgCLI)
        {
            byte CodeCLI;
            uint lCampo;
            string OraCorrente = DateTime.Now.ToString("HH:mm:ss"); // Ora di sistema
            myFtdiDevice.SetDTR(false); //Disattiva MT88E39
            switch (StatoMSg) {
                case 0:
                    /*
                    string msgMon = string.Empty;
                    foreach (byte b in MsgCLI) msgMon += String.Format("{0:X2} ", b); //Per ogni byte in msg (byte[]) genera la stringa msgMon
                    ScriviSuMonitor(msgMon);
                    */
                    uint i = 0;
                    do {
                        CodeCLI = MsgCLI[i++];
                        switch (CodeCLI) {
                            case 0x01: //Campo Data e Ora
                                lCampo = MsgCLI[i++];
                                
                                i += lCampo;
                                break;
                            case 0x02: //Campo Numero di telefono
                                lCampo = MsgCLI[i++];
                                char[] nTel = new char[lCampo];
                                for (int k = 0; k < lCampo; k++) nTel[k] = (char)MsgCLI[i++];
                                String NumTel = new string(nTel);

                                // When a program often has to try keys that turn out not to
                                // be in the dictionary, TryGetValue can be a more efficient 
                                // way to retrieve values.
                                string value = "";
                                if (rubrica.TryGetValue(NumTel, out value))
                                {
                                    Annuncio = "Chiamata in arrivo da,, " + value;
                                    synth.SpeakAsync(Annuncio);
                                    ScriviSuMonitor("Ore: " + OraCorrente);
                                    ScriviSuMonitor("Numero chiamante: " + NumTel + " : " + value);
                                }
                                else
                                {
                                    Annuncio = "Chiamata dal numero " + NumTel;
                                    synth.SpeakAsync(Annuncio);
                                    ScriviSuMonitor("Ore: " + OraCorrente);
                                    ScriviSuMonitor("Numero chiamante: " + NumTel);
                                }
                                break;
                            case 0x04: //Anonimato
                                lCampo = MsgCLI[i++];
                                if (MsgCLI[i++] == 0x50)
                                {
                                    synth.SpeakAsync("Chiamata in arrivo da numero anonimo, al quarto squillo verrà rifiutata");
                                    ScriviSuMonitor("Ore: " + OraCorrente);
                                    ScriviSuMonitor("Chiamata in arrivo da numero anonimo");
                                    ChiamataAnonima = true;
                                }
                                break;
                        }
                    } while (i < MsgCLI.Length);
                    break;
                case 1:
                    ScriviSuMonitor("Ore: " + OraCorrente);
                    ScriviSuMonitor("Errore di CheckSum sui dati CLI");
                    synth.SpeakAsync("Errore di lettura dei dati del chiamante");
                    break;
                case 2:
                    ScriviSuMonitor("Ore: " + OraCorrente);
                    ScriviSuMonitor("Mancanza di CLI");
                    synth.SpeakAsync("Errore di lettura dei dati del chiamante");
                    break;
                case 3:
                    ScriviSuMonitor("Numero squilli: " + MsgCLI[0]);
                   break;
            }
        }


        //-------------------------------------------------------------------------------------------------------------------------------------------
        // Scadenza del timer tmrFTDI per la ricerca di FT232RL (usato per creare il servizio windows)
        //-------------------------------------------------------------------------------------------------------------------------------------------
        public void tmrFTDI_Tick(object sender, EventArgs e)
        {
            tmrFTDI.Enabled = false;
            string DescrizoneFTDI = "FT232R USB UART";
            //ftStatus = OpenFTDI_port(UsbPort);  //Apertura FTDI collegato su una specifica porta USB
            ftStatus = OpenFTDI_dsc(DescrizoneFTDI);
            if (ftStatus == FTDI.FT_STATUS.FT_OK) {
                synth.SpeakAsync("FTDI Collegato");
                ScriviSuMonitor(DescrizoneFTDI + " Collegato");
            } else {
                synth.SpeakAsync("Collegare FTDI");
                ScriviSuMonitor("Collegare FTDI con descrizione " + DescrizoneFTDI);
                tmrFTDI.Enabled = true;
            }
        }

        //-------------------------------------------------------------------------------------------------------------------------------------------
        // Tenta di aprire il dispositivo FTDI corrispondente alla descrizione
        //-------------------------------------------------------------------------------------------------------------------------------------------
        private FTDI.FT_STATUS OpenFTDI_dsc(string Descrizione)
        {
            ftStatus = myFtdiDevice.OpenByDescription(Descrizione);
            if (ftStatus == FTDI.FT_STATUS.FT_OK)
            {
                myFtdiDevice.SetDTR(false);

                ftStatus = myFtdiDevice.SetBaudRate(1200);
                if (ftStatus != FTDI.FT_STATUS.FT_OK) return ftStatus;
                // Set data characteristics - 8n1
                ftStatus = myFtdiDevice.SetDataCharacteristics(FTDI.FT_DATA_BITS.FT_BITS_8, FTDI.FT_STOP_BITS.FT_STOP_BITS_1, FTDI.FT_PARITY.FT_PARITY_NONE);
                if (ftStatus != FTDI.FT_STATUS.FT_OK) return ftStatus;
                // Set flow control - None
                ftStatus = myFtdiDevice.SetFlowControl(FTDI.FT_FLOW_CONTROL.FT_FLOW_NONE, 0x11, 0x13);
                if (ftStatus != FTDI.FT_STATUS.FT_OK) return ftStatus;
                // Set read timeout to 5 seconds, write timeout to infinite
                ftStatus = myFtdiDevice.SetTimeouts(5000, 0);
                if (ftStatus != FTDI.FT_STATUS.FT_OK) return ftStatus;

                // Creazione di un thread separato che resta in attesa di uno squillo e dei successivi dati seriali da FTDI
                BackgroundWorker dataReceivedHandler = new BackgroundWorker();
                dataReceivedHandler.DoWork += ReadData;
                if (!dataReceivedHandler.IsBusy) dataReceivedHandler.RunWorkerAsync();
            }
            return ftStatus;
        }


        //-------------------------------------------------------------------------------------------------------------------------------------------
        // Tenta di aprire il dispositivo FTDI che si trova sulla porta USB 'LocPortaUSB' (Hub*16 + Port)
        //-------------------------------------------------------------------------------------------------------------------------------------------
        private FTDI.FT_STATUS OpenFTDI_port(uint LocPortaUSB)
        {
            ftStatus = myFtdiDevice.OpenByLocation(LocPortaUSB);
            if (ftStatus == FTDI.FT_STATUS.FT_OK)
            {
                myFtdiDevice.SetDTR(false);

                ftStatus = myFtdiDevice.SetBaudRate(1200);
                if (ftStatus != FTDI.FT_STATUS.FT_OK) return ftStatus;
                // Set data characteristics - 8n1
                ftStatus = myFtdiDevice.SetDataCharacteristics(FTDI.FT_DATA_BITS.FT_BITS_8, FTDI.FT_STOP_BITS.FT_STOP_BITS_1, FTDI.FT_PARITY.FT_PARITY_NONE);
                if (ftStatus != FTDI.FT_STATUS.FT_OK) return ftStatus;
                // Set flow control - None
                ftStatus = myFtdiDevice.SetFlowControl(FTDI.FT_FLOW_CONTROL.FT_FLOW_NONE, 0x11, 0x13);
                if (ftStatus != FTDI.FT_STATUS.FT_OK) return ftStatus;
                // Set read timeout to 5 seconds, write timeout to infinite
                ftStatus = myFtdiDevice.SetTimeouts(5000, 0);
                if (ftStatus != FTDI.FT_STATUS.FT_OK) return ftStatus;

                // Creazione di un thread separato che resta in attesa di uno squillo e dei successivi dati seriali da FTDI
                BackgroundWorker dataReceivedHandler = new BackgroundWorker();
                dataReceivedHandler.DoWork += ReadData;
                if (!dataReceivedHandler.IsBusy) dataReceivedHandler.RunWorkerAsync();
            }
            return ftStatus;
        }




        //-------------------------------------------------------------------------------------------------------------------------------------------
        // Form Load
        //-------------------------------------------------------------------------------------------------------------------------------------------
        private void Form1_Load(object sender, EventArgs e)
        {
            // Configure the audio output. 
            synth.SetOutputToDefaultAudioDevice();
            // Regola il volume
            synth.Volume = 50; //da 0 a 100
            // Speak a string asynchronously.
            synth.SpeakAsync("Rilevatore Chiamata Avviato");

            try
            {
                using (StreamReader sr = new StreamReader("rubrica.txt"))
                {
                    while (true)
                    {
                        // Legge una linea del file testo *.hex alla volta
                        String line = sr.ReadLine();
                        if (line == null || line == "") break;
                        string Nome = line.Substring(0, line.IndexOf(','));
                        string nTel = line.Substring(line.IndexOf(',') + 1);
                        rubrica.Add(nTel, Nome);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Errore lettura file rubrica: " + ex.Message);
            }

            tmrFTDI = new System.Windows.Forms.Timer();
            tmrFTDI.Interval = 4000;
            tmrFTDI.Tick += new System.EventHandler(this.tmrFTDI_Tick);
            tmrFTDI.Start();
            myFtdiDevice = new FTDI();
        }

        //-------------------------------------------------------------------------------------------------------------------------------------------
        // Scrivi su monitor
        //-------------------------------------------------------------------------------------------------------------------------------------------
        void ScriviSuMonitor(string inMsg)
        {
            string msg = "";
            lboxMonitor.Items.Add(msg + inMsg); //Addiziona la stringa al controllo ListBox
            if (lboxMonitor.Items.Count == 50) lboxMonitor.Items.RemoveAt(0); //Limita listbox a 50
            lboxMonitor.SelectedIndex = lboxMonitor.Items.Count - 1; //Seleziona e quindi visualizza sempre l'ultimo con evidenziazione
                                                                     //Monitor.TopIndex = Monitor.Items.Count-1; //Visualizza sempre l'ultimo usando la proprietà TopIndex senza evidenziazione
            lboxMonitor.Refresh();
        }

    }
    

}
